# define context class
# written by MOROHOSHI Akihiko <moro@nii.ac.jp>

########################################################################
# Context
class Context
    attr_accessor :bold, :emphasis, :strong_emphasis,
	:italic, :typewriter
    attr_reader :underline, :size, :color

    SIZE_UNKNOWN = -1
    COLOR_UNKNOWN = ?U

    def initialize
	@bold = @emphasis = @strong_emphasis = @italic = @typewriter =false
	@underline = 0
	@size = SIZE_UNKNOWN
	@color = COLOR_UNKNOWN
    end

    def color=(c)
	case c
	when ?r, ?g, ?b, ?c, ?m, ?y, ?k, ?w
	    @color = c
	else
	    raise "unknown color: #{c.inspect}"
	end
    end
    def size=(s)
	if (0..9).include? s
	    @size = s
	else
	    raise "invalid size: #{s}"
	end
    end

    def underline_add
	if @underline < 3
	    @underline += 1
	end
    end

    def underline_reset
	@underline = 0
    end

    def space_context()
	c = self.dup
	c.bold = c.emphasis = c.strong_emphasis = c.italic = false
	c.color = ?w if c.underline == 0
	c
    end

    def ==(other)
	return false if other.type != self.type

	return false if other.bold != @bold
	return false if other.typewriter != @typewriter
	return false if other.italic != @italic
	return false if other.underline != @underline
	return false if other.color != @color
	return false if other.size != @size
	return false if other.strong_emphasis != @strong_emphasis
	return false if (!@strong_emphasis) and other.emphasis != @emphasis

	true
    end

    def distance(other)
	d = 0
	d += 1 if other.bold != @bold
	d += 1 if other.typewriter != @typewriter
	d += 1 if other.italic != @italic
	d += (other.underline - @underline).abs
	d += 1 if other.color != @color
	d += 1 if other.size != @size
	d += 1 if other.strong_emphasis != @strong_emphasis
	unless (other.strong_emphasis == @strong_emphasis and
		@strong_emphasis)
	    d += 1 if other.emphasis != @emphasis
	end
	
	d
    end

    def distance_min_est(other)
	d = 0
	d1 = 0
	d1 += 1 if other.bold != @bold
	d1 += 1 if other.typewriter != @typewriter
	d1 += 1 if other.italic != @italic
	d1 = 1 if d1 > 0
	d += d1
	
	d2 = (other.underline - @underline).abs
	if d2 > 0 and @underline == 0
	    d2 = (d1 > 0 ? 0 : 1)
	end
	d += d2
	
	d += 1 if other.color != @color
	d += 1 if other.size != @size
	d += 1 if other.strong_emphasis != @strong_emphasis
	unless (other.strong_emphasis == @strong_emphasis and
		@strong_emphasis)
	    d += 1 if other.emphasis != @emphasis
	end
	
	d
    end
end

########################################################################
# utility
def context_modified_by_tag(context,tag)
    cont = context.dup
    case tag
    when 'B'
	cont.bold = true
    when 'EM'
	cont.emphasis = (! cont.emphasis)
    when 'S'
	cont.strong_emphasis = true
    when 'I'
	cont.italic = true
    when 'PL'
	cont.underline_reset
	cont.bold = false
	cont.emphasis = false
	cont.italic = false
	cont.strong_emphasis = false
	cont.typewriter = false
    when 'TT'
	cont.typewriter = true
    when 'U'
	cont.underline_add
    when /\A\d\Z/
	cont.size = tag.to_i
    when /\A[rgbcmykw]\Z/
	cont.color = tag[0]
    else
	raise "unknown tag: #{tag}"
    end
    cont
end

